open Defs
open Printf


type value = 
    VInteger of int
  | VReal of float
  | VBoolean of bool
  | VUndef

let valuetable = Hashtbl.create 10007

let assign_value ((s:string),(v:value)) = Hashtbl.add valuetable s v

let lookup_value (inf,s) = 
  try Hashtbl.find valuetable s
  with Not_found -> let r = print_Range inf in
                    printf "%s: value of %s undefined\n" r s;
                    VUndef

let mkReal x = match x with
  VInteger i -> float_of_int i
| VReal r -> r
| _ -> Printf.printf "error in conversion to real"; exit(0)

let rec interProgram (Program (_,_,sequence)) = interSeq sequence
and interSeq xs = List.iter interInfoStmt xs
and interInfoStmt (inf,s) = match s with
  Assign (s,e) -> let v = interInfoExpr e in
                  let t = (gI e).typ in
                  assign_value (s, if t=TReal then VReal (mkReal v) else v)
| Read s -> 
    begin
      match lookup_type (inf,s) with
	TInteger -> 
	  Printf.printf "integer     %s <- " s;
          let i = read_int() in
          assign_value (s, VInteger i)
      | TReal -> 
	  Printf.printf "real number %s <- " s;
          flush stdout;
          let r = read_float() in          
          assign_value (s, VReal r)
      | _ -> Printf.printf "input type unknown of %s" s
    end
| Write e -> 
    begin
      match interInfoExpr e with
        VInteger i -> Printf.printf "integer       -> %d\n" i
      | VReal r ->    Printf.printf "real number   -> %f\n" r
      | VBoolean true  ->    Printf.printf "boolean value   -> true\n"
      | VBoolean false ->    Printf.printf "boolean value   -> false\n"
    end
| Sequence xs -> interSeq xs
| IfThenElse (cond,cthen,celse) ->
    begin
      match (interInfoExpr cond , celse) with
	(VBoolean true,_) -> interInfoStmt cthen
      | (VBoolean false, Some stmt) -> interInfoStmt stmt
      | _ -> ()
    end 
| While (cond,body) ->
    begin
      match interInfoExpr cond with
	VBoolean true -> interInfoStmt body; interInfoStmt (inf,While (cond,body))
      | _ -> ()
    end
| For (var,estart,eend,vstep,body) ->
    let vcurr = ref (match interInfoExpr estart with VInteger i -> i) in
    let VInteger vend = interInfoExpr eend in
    while (if vstep>0 then !vcurr<=vend else !vcurr>=vend) do
      assign_value(var,VInteger !vcurr);
      interInfoStmt body;
      if vstep>0 then incr vcurr else decr vcurr
    done
and interInfoExpr (inf,e) = match e with
  CInteger i -> VInteger i
| CReal r -> VReal r
| CBoolean b -> VBoolean b
| Var s -> lookup_value (inf,s)
| Par e -> interInfoExpr e
| Op (op,xs) -> interOp (gE op) xs
and interOp op xs = let ys = List.map interInfoExpr xs in
                    applyOp op ys
and unB (VBoolean b) = b
and applyOp op xs = match (op,xs) with
   (OAdd, [VInteger a; VInteger b]) -> VInteger (a+b)
|  (OAdd, [a;b]) -> VReal (mkReal a +. mkReal b)
|  (OSub, [VInteger a; VInteger b]) -> VInteger (a-b)
|  (OSub, [a;b]) -> VReal (mkReal a -. mkReal b)
|  (OMul, [VInteger a; VInteger b]) -> VInteger (a*b)
|  (OMul, [a;b]) -> VReal (mkReal a *. mkReal b)
|  (OFrac, [a;b]) -> VReal (mkReal a /. mkReal b)
|  (ODiv, [VInteger a; VInteger b]) -> VInteger (a/b)
|  (OMod, [VInteger a; VInteger b]) -> VInteger (a mod b)
|  (OPos, [VInteger a]) -> VInteger a
|  (OPos, [VReal a]) -> VReal a
|  (ONeg, [VInteger a]) -> VInteger (-a)
|  (ONeg, [VReal a]) -> VReal (-. a)
|  (ONot, [VBoolean a]) -> VBoolean (not a)
|  (OAnd, [VBoolean a; VBoolean b]) -> VBoolean (a && b)
|  (OOr, [VBoolean a; VBoolean b]) -> VBoolean (a || b)
|  (OEq, [VInteger a; VInteger b]) -> VBoolean (a=b)
|  (OEq, [VInteger a; VReal b]) -> VBoolean (float_of_int a = b)
|  (OEq, [VReal a; VInteger b])  -> VBoolean (a = float_of_int b)
|  (OEq, [VReal a; VReal b]) -> VBoolean (a=b)
|  (OEq, [VBoolean a; VBoolean b]) -> VBoolean (a=b)
|  (OLess, [VInteger a; VInteger b]) -> VBoolean (a<b)
|  (OLess, [VInteger a; VReal b]) -> VBoolean (float_of_int a < b)
|  (OLess, [VReal a; VInteger b]) -> VBoolean (a < float_of_int b)
|  (OLess, [VReal a; VReal b]) -> VBoolean (a<b)
|  (OLess, [VBoolean a; VBoolean b]) -> VBoolean (a<b)
|  (ONeq, [x;y]) -> VBoolean (not (unB (applyOp OEq [x;y])))
|  (OLeq, [x;y]) 
   -> VBoolean (     (unB (applyOp OEq   [x;y]))
                  || (unB (applyOp OLess [x;y])) )
|  (OGeq, [x;y]) -> VBoolean (not (unB (applyOp OLess [x;y])))
|  (OGrea, [x;y]) -> VBoolean (not (unB (applyOp OLeq  [x;y])))


