{
  open Defs          
  open Parser        (* The type token is defined in parser.mli *)
  exception Eof

  let keyword_table = Hashtbl.create 53
  let _ =
    List.iter (fun (kwd, tok) -> Hashtbl.add keyword_table kwd tok)
              [ ("for",    fun x -> TFOR x);
                ("to",     fun x -> TTO x);
                ("downto",  fun x -> TDOWNTO x);
                ("do",   fun x -> TDO x);
                ("while",   fun x -> TWHILE x);
                ("if",    fun x -> TIF x);
                ("then",  fun x -> TTHEN x);
                ("else",  fun x -> TELSE x);
                ("not",   fun x -> TNOT x);
                ("and",   fun x -> TAND x);
                ("or",    fun x -> TOR x);
                ("div",   fun x -> TDIV x);
                ("mod",   fun x -> TMOD x);
                ("read",  fun x -> TREAD x);
                ("write", fun x -> TWRITE x);
                ("begin", fun x -> TBEGIN x);
                ("end",   fun x -> TEND x);
                ("program", fun x -> TPROGRAM x);
                ("var",   fun x -> TVAR x);
                ("integer", fun x -> TTYPECONST (x,TInteger));
                ("real",    fun x -> TTYPECONST (x,TReal));
                ("boolean", fun x -> TTYPECONST (x,TBoolean));
                ("false", fun x -> TBOOLEAN (x,false));
                ("true",  fun x -> TBOOLEAN (x,true))
              ]

  let curr_line = ref 1
  let curr_col  = ref 0
  let curr_tok  = ref ""
  let nested_comm = ref 0

  let mr (ln,newpos,len) = { locL={line=ln;col=newpos-len};
                             locR={line=ln;col=newpos-1};
                             typ=TUndef }

  let tok s = begin
               let n = String.length s in 
               curr_col := !curr_col + n;
               curr_tok := s;
               (mr (!curr_line,!curr_col,n))
              end
}
  
let digit  = ['0'-'9']           
let letter = ['a'-'z''A'-'Z']
let linteger = digit+
let lreal = digit+('.'digit+)?(['e''E'](['+''-']?)digit+)?
let name   = letter(letter|digit)*
    
rule comment = parse
  "*)"            { curr_col := !curr_col + 2;
                    decr nested_comm; 
                    if !nested_comm = 0 then token lexbuf else comment lexbuf}
| "(*"            { curr_col := !curr_col + 2; incr nested_comm; comment lexbuf }
| ['\t']          { curr_col := (let p = !curr_col in p+8 - p mod 8); token lexbuf } 
| ['\n']          { curr_col := 0; incr curr_line; comment lexbuf }
| _               { incr curr_col; comment lexbuf }
and token = parse
  "(*"            { curr_col := !curr_col + 2; incr nested_comm; comment lexbuf }
| [' ']           { incr curr_col; token lexbuf }
| ['\t']          { curr_col := (let p = !curr_col in p+8 - p mod 8); token lexbuf } 
| ['\n']          { curr_col := 0; incr curr_line; token lexbuf }
| linteger as lxm { TINTEGER (tok lxm, int_of_string lxm) }
| lreal    as lxm { TREAL    (tok lxm, float_of_string lxm) }
| name     as lxm { let rng = tok lxm in 
                    try Hashtbl.find keyword_table (String.lowercase lxm) rng 
                    with Not_found -> TIDENT (rng,lxm) }
| '+'             { TPLUS      (tok "+") }
| '-'             { TMINUS     (tok "-") }
| '*'             { TTIMES     (tok "*") }
| '/'             { TSLASH     (tok "/") }
| '('             { TLPAREN    (tok "(") }
| ')'             { TRPAREN    (tok ")") }
| ';'             { TSEMICOLON (tok ";") }
| ':'             { TCOLON     (tok ":") }
| ','             { TCOMMA     (tok ",") }
| '.'             { TDOT       (tok ".") }
| ":="            { TASSIGN    (tok ":=") }
| "="             { TEQ        (tok "=") }
| "<>"            { TNEQ       (tok "<>") }
| "<"             { TLESS      (tok "<") }
| "<="            { TLEQ       (tok "<=") }
| ">"             { TGREA      (tok ">") }
| ">="            { TGEQ       (tok ">=") }
| eof             { raise Eof }
